/**
 ** @file mod_xml.c
 ** libxml2 module (See http://xmlsoft.org)
 **
 ** @author Baptiste GOURDIN <gourdin@lsv.ens-cachan.fr>
 ** @version 0.1
 ** @ingroup modules
 **
 **
 ** @date  Started on:  Tue Apr 12 11:20:38 CEST 2011
 **/

/*
 * See end of file for LICENSE and COPYRIGHT informations.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif


#include <string.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <libxml/xmlschemas.h>
#include <time.h>

#include "orchids.h"
#include "mod_xml.h"


input_module_t mod_xml;

xmlNodePtr
xml_walk_and_create (xmlXPathContextPtr ctx,
		     xmlChar	*path)
{
  xmlXPathObjectPtr	xpathObj = NULL;
  xmlNodePtr		node = NULL;
  unsigned int		path_len = 0;
  unsigned int		new_len = 0;
  xmlChar		*off = NULL;

  if (xmlXPathCompile(path) == NULL)
  {
    DebugLog(DF_MOD, DS_ERROR,
	     "Error: unable to evaluate xpath expression \"%s\"\n", path);
    return NULL;
  }

  path_len = xmlStrlen(path);

  do {
    xpathObj = xmlXPathEvalExpression(path, ctx);
    if ((xpathObj == NULL) || (xpathObj->nodesetval == 0)
	|| (xpathObj->nodesetval->nodeNr == 0))
    {
      off = (xmlChar*)strrchr((char*)path, '/');
      if (off != NULL)
	*off = 0;
    }
  } while (((xpathObj == NULL) || (xpathObj->nodesetval == 0)
	    || (xpathObj->nodesetval->nodeNr == 0)) && (off != path));

  if (off != path)
    node = xpathObj->nodesetval->nodeTab[0];
  else
  {
    DebugLog(DF_MOD, DS_ERROR,
	     "Error: wrong root node \"%s\"\n", path);
    return NULL;
  }

  new_len = xmlStrlen (path);
  while  (new_len != path_len)
  {
    path[new_len] = '/';
    node = xmlNewChild(node, NULL, path + new_len + 1, NULL);
    new_len = xmlStrlen (path);
  }

  xmlXPathFreeObject(xpathObj);

  return node;
}

char*
xml_get_string(xml_doc_t	*xml_doc,
	       char		*path)
{
  xmlChar	*text;
  xmlXPathObjectPtr	xpathObj = NULL;
  xmlNodePtr		node = NULL;

  xpathObj = xmlXPathEvalExpression(BAD_CAST path, xml_doc->xpath_ctx);
  if ((xpathObj == NULL) || (xpathObj->nodesetval == 0)
      || (xpathObj->nodesetval->nodeNr == 0))
  {
    DebugLog(DF_ENG, DS_ERROR, "error : no nodes for path %s\n", path);
    return (NULL);
  }
  else
  {
    node = xpathObj->nodesetval->nodeTab[0];
    if ((text = xmlNodeGetContent(node)) == NULL)
      return (NULL);
    else
      return ((char*)text);
  }
}

static void
issdl_xml_get_string(orchids_t *ctx, state_instance_t *state)
{
  ovm_var_t	*var1;
  ovm_var_t	*var2;
  xml_doc_t	*xml_doc;
  xmlXPathObjectPtr	xpathObj = NULL;
  xmlNodePtr		node = NULL;
  ovm_var_t	*res;
  size_t	res_len;
  xmlChar	*text;

  var1 = stack_pop(ctx->ovm_stack);
  var2 = stack_pop(ctx->ovm_stack);

  if (!var1 || !var2 || (TYPE(var1) != T_PTR32) || (TYPE(var2) != T_STR)) {
    DebugLog(DF_ENG, DS_ERROR, "parameter error\n");
    return ;
  }

  if ((xml_doc = PTR32(var1)) == NULL)  {
    DebugLog(DF_ENG, DS_ERROR, "error : pointer is null\n");
    return ;
  }

  xpathObj = xmlXPathEvalExpression(BAD_CAST STR(var2), xml_doc->xpath_ctx);
  if ((xpathObj == NULL) || (xpathObj->nodesetval == 0)
      || (xpathObj->nodesetval->nodeNr == 0))
  {
    DebugLog(DF_ENG, DS_ERROR, "error : no nodes for path %s\n", STR(var2));
    stack_push(ctx->ovm_stack, ovm_str_new(0));
  }
  else
  {
    node = xpathObj->nodesetval->nodeTab[0];
    if ((text = xmlNodeGetContent(node)) == NULL)
      stack_push(ctx->ovm_stack, ovm_str_new(0));
    else
    {
      res_len = strlen((char*)text);
      res = ovm_str_new(res_len);
      memcpy (STR(res), text, res_len);
      stack_push(ctx->ovm_stack, res);
    }
  }
}

static void
print_schema_validity_error (void* ctx, const char* msg, ...)
{
  va_list ap;

  va_start(ap, msg);
  DebugLogVa(DF_MOD, DS_ERROR, msg, ap);
  va_end(ap);
}

static void
print_schema_validity_warning (void* ctx, const char* msg, ...)
{
  va_list ap;

  va_start(ap, msg);
  DebugLogVa(DF_MOD, DS_WARN, msg, ap);
  va_end(ap);
}

xmlNode*
new_xs_datetime_node(xmlNode	*parent,
		     char	*name,
		     struct tm *time)
{
  char		buff[32];
  size_t	len;

  len = strftime(buff, sizeof (buff), "%Y-%m-%dT%H:%M:%S%z", time);

  if ((buff[len - 5] == '+') ||
      (buff[len - 5] == '-'))
  {
    buff[len + 1] = 0;
    buff[len] = buff[len - 1];
    buff[len - 1] = buff[len - 2];
    buff[len - 2] = ':';
  }

  return xmlNewChild(parent, NULL, BAD_CAST name,
		     BAD_CAST strdup(buff));
}

char
validate_xml_with_schema(xmlSchemaPtr	schema,
			 char		*doc)
{
  xmlSchemaValidCtxtPtr		schema_valid = NULL;
  int				res= 0;

  if ((schema_valid = xmlSchemaNewValidCtxt (schema)) == NULL)
  {
    DebugLog(DF_MOD, DS_ERROR, "unable to create a SchemaValidCtx\n");
    return 1;
  }

  xmlSchemaSetValidErrors(schema_valid,
  			  print_schema_validity_error,
  			  print_schema_validity_warning,
  			  NULL);

  res = xmlSchemaValidateFile (schema_valid, doc, 1);
  xmlSchemaFreeValidCtxt(schema_valid);

  if (res == 0)
    return 1;
  else if (res == -1)
    DebugLog(DF_MOD, DS_ERROR, "libxml internal error \n");
  else
    DebugLog(DF_MOD, DS_WARN, "xml structure not valid\n");
  return 0;
}


static void *
mod_xml_preconfig(orchids_t *ctx, mod_entry_t *mod)
{
  DebugLog(DF_MOD, DS_INFO, "load() mod_xml@%p\n", (void *) &mod_xml);


  register_lang_function(ctx,
                         issdl_xml_get_string,
                         "xml_get_str", 2,
                         "get a string from an xml doc using an xpath request");


  return (NULL);
}


input_module_t mod_xml = {
  MOD_MAGIC,
  ORCHIDS_VERSION,
  "xml",
  "CeCILL2",
  NULL,
  NULL,
  mod_xml_preconfig,
  NULL,
  NULL
};

/*
** Copyright (c) 2002-2005 by Julien OLIVAIN, Laboratoire Spécification
** et Vérification (LSV), CNRS UMR 8643 & ENS Cachan.
**
** Julien OLIVAIN <julien.olivain@lsv.ens-cachan.fr>
**
** This software is a computer program whose purpose is to detect intrusions
** in a computer network.
**
** This software is governed by the CeCILL license under French law and
** abiding by the rules of distribution of free software.  You can use,
** modify and/or redistribute the software under the terms of the CeCILL
** license as circulated by CEA, CNRS and INRIA at the following URL
** "http://www.cecill.info".
**
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
**
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
**
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL license and that you accept its terms.
*/


/* End-of-file */
